package dev.flyfish.framework.repository;

import dev.flyfish.framework.repository.base.DomainRepository;
import org.reactivestreams.Publisher;
import org.springframework.data.repository.NoRepositoryBean;
import org.springframework.data.repository.reactive.ReactiveCrudRepository;
import org.springframework.lang.NonNull;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;


/**
 * 默认的持久层dao
 *
 * @param <T> 泛型
 */
@NoRepositoryBean
public interface DefaultReactiveRepository<T> extends ReactiveCrudRepository<T, String>, ReactiveQueryModelExecutor<T>, DomainRepository<T> {

    /**
     * 插入给定的实体. 假设该实体是新的，以便应用插入优化。
     * 使用返回的实例进行后续操作，因为保存操作可能会完全改变实体实例。
     * 建议使用 {@link #save(Object)} 来避免使用特定存储的 API。
     *
     * @param entity 要插入的实体，不能为空
     * @return 保存后的实体
     */
    @NonNull
    <S extends T> Mono<S> insert(@NonNull S entity);

    /**
     * 插入给定的实体列表. 假设这些实体是新的，以便应用插入优化。
     * 使用返回的实例进行后续操作，因为保存操作可能会完全改变实体实例。
     * 建议使用 {@link #save(Object)} 来避免使用特定存储的 API。
     *
     * @param entities 实体列表，不能为空
     * @return 保存后的实体列表
     */
    <S extends T> Flux<S> insert(Iterable<S> entities);

    /**
     * 插入给定的实体发布者发布的实体. 假设这些实体是新的，以便应用插入优化。
     * 使用返回的实例进行后续操作，因为保存操作可能会完全改变实体实例。
     * 建议使用 {@link #save(Object)} 来避免使用特定存储的 API。
     *
     * @param entities 发布的实体列表，不能为空
     * @return 保存后的实体列表
     */
    <S extends T> Flux<S> insert(Publisher<S> entities);
}
