package dev.flyfish.framework.r2dbc.config.mapping;

import dev.flyfish.framework.utils.JacksonUtil;
import io.r2dbc.spi.Readable;
import io.r2dbc.spi.ReadableMetadata;
import io.r2dbc.spi.Row;
import io.r2dbc.spi.RowMetadata;
import org.springframework.data.r2dbc.convert.R2dbcConverter;
import org.springframework.data.r2dbc.core.DefaultReactiveDataAccessStrategy;
import org.springframework.data.r2dbc.core.PredefinedStatementMapper;
import org.springframework.data.r2dbc.core.ReactiveDataAccessStrategy;
import org.springframework.data.r2dbc.core.StatementMapper;
import org.springframework.data.r2dbc.dialect.R2dbcDialect;
import org.springframework.data.r2dbc.mapping.OutboundRow;
import org.springframework.data.r2dbc.query.FullUpdateMapper;
import org.springframework.data.r2dbc.query.UpdateMapper;
import org.springframework.data.relational.core.dialect.RenderContextFactory;
import org.springframework.data.relational.core.mapping.RelationalPersistentEntity;
import org.springframework.data.relational.core.mapping.RelationalPersistentProperty;
import org.springframework.data.relational.core.sql.SqlIdentifier;
import org.springframework.data.relational.core.sql.render.RenderContext;
import org.springframework.data.relational.domain.RowDocument;
import org.springframework.lang.NonNull;
import org.springframework.r2dbc.core.Parameter;
import org.springframework.r2dbc.core.PreparedOperation;
import org.springframework.util.Assert;
import org.springframework.util.ClassUtils;

import java.util.Collection;
import java.util.List;
import java.util.function.BiFunction;

/**
 * 全支持的数据访问策略，将数组扁平化转换，方便支持json
 *
 * @author wangyu
 */
@SuppressWarnings("deprecation")
public class FullReactiveDataAccessStrategy implements ReactiveDataAccessStrategy {

    private final DefaultReactiveDataAccessStrategy delegated;

    private final R2dbcConverter converter;

    private final UpdateMapper updateMapper;

    private final StatementMapper statementMapper;

    /**
     * 最大化配置的构造器，仅提供该实现
     *
     * @param dialect   别名实现
     * @param converter 转换器
     */
    public FullReactiveDataAccessStrategy(R2dbcDialect dialect, R2dbcConverter converter) {
        Assert.notNull(dialect, "Dialect must not be null");
        Assert.notNull(converter, "RelationalConverter must not be null");
        this.delegated = new DefaultReactiveDataAccessStrategy(dialect, converter);
        RenderContextFactory factory = new RenderContextFactory(dialect);
        RenderContext renderContext = factory.createRenderContext();

        this.converter = converter;
        this.updateMapper = new FullUpdateMapper(dialect, converter, renderContext);

        this.statementMapper = new PredefinedStatementMapper(dialect, renderContext, this.updateMapper,
                delegated.getMappingContext());
    }

    @Override
    @NonNull
    public List<SqlIdentifier> getAllColumns(@NonNull Class<?> entityType) {
        return delegated.getAllColumns(entityType);
    }

    @Override
    @NonNull
    public List<SqlIdentifier> getIdentifierColumns(@NonNull Class<?> entityType) {
        return delegated.getIdentifierColumns(entityType);
    }

    @Override
    @NonNull
    public OutboundRow getOutboundRow(@NonNull Object object) {
        Assert.notNull(object, "Entity object must not be null");

        OutboundRow row = new OutboundRow();

        converter.write(object, row);

        RelationalPersistentEntity<?> entity = getRequiredPersistentEntity(ClassUtils.getUserClass(object));

        for (RelationalPersistentProperty property : entity) {
            Parameter value = row.get(property.getColumnName());
            if (value != null && shouldConvertArrayValue(property, value)) {
                Parameter writeValue = getArrayValue(value);
                row.put(property.getColumnName(), writeValue);
            }
        }

        return row;
    }

    @Override
    @NonNull
    public Parameter getBindValue(@NonNull Parameter value) {
        return this.updateMapper.getBindValue(value);
    }

    @Override
    @NonNull
    public <T> BiFunction<Row, RowMetadata, T> getRowMapper(@NonNull Class<T> typeToRead) {
        return delegated.getRowMapper(typeToRead);
    }

    @Override
    @NonNull
    public RowDocument toRowDocument(@NonNull Class<?> type, @NonNull Readable row, @NonNull Iterable<? extends ReadableMetadata> metadata) {
        return delegated.toRowDocument(type, row, metadata);
    }

    @Override
    @NonNull
    public SqlIdentifier getTableName(@NonNull Class<?> type) {
        return delegated.getTableName(type);
    }

    @Override
    @NonNull
    public PreparedOperation<?> processNamedParameters(@NonNull String query, @NonNull NamedParameterProvider parameterProvider) {
        return delegated.processNamedParameters(query, parameterProvider);
    }

    private Parameter getArrayValue(Parameter value) {
        if (value.getType().equals(byte[].class)) {
            return value;
        }
        return Parameter.fromOrEmpty(JacksonUtil.toJson(value.getValue()).orElse(null), String.class);
    }

    @Override
    @NonNull
    public String toSql(@NonNull SqlIdentifier identifier) {
        return this.updateMapper.toSql(identifier);
    }

    @Override
    @NonNull
    public StatementMapper getStatementMapper() {
        return this.statementMapper;
    }

    @Override
    @NonNull
    public R2dbcConverter getConverter() {
        return converter;
    }

    /**
     * 判断是否需要转换数据值
     *
     * @param property 属性
     * @param value    值表示
     * @return 结果
     */
    private boolean shouldConvertArrayValue(RelationalPersistentProperty property, Parameter value) {
        if (!property.isCollectionLike()) {
            return false;
        }
        if (value.hasValue() && (value.getValue() instanceof Collection || value.getValue().getClass().isArray())) {
            return true;
        }
        if (Collection.class.isAssignableFrom(value.getType()) || value.getType().isArray()) {
            return true;
        }
        return false;
    }

    @SuppressWarnings("unchecked")
    private RelationalPersistentEntity<?> getRequiredPersistentEntity(Class<?> typeToRead) {
        return converter.getMappingContext().getRequiredPersistentEntity(typeToRead);
    }

}
