package dev.flyfish.framework.user.config;

import dev.flyfish.framework.user.domain.AdminUserDetails;
import dev.flyfish.framework.domain.po.User;
import dev.flyfish.framework.enums.UserStatus;
import dev.flyfish.framework.service.AuthenticationAuditor;
import dev.flyfish.framework.service.AuthenticationLogger;
import dev.flyfish.framework.user.service.UserService;
import dev.flyfish.framework.transform.ResultDataTransformer;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.web.server.WebFilterExchange;
import reactor.core.publisher.Mono;

import jakarta.annotation.Resource;
import java.time.LocalDateTime;

/**
 * 提供对象写入，登录登出审计
 *
 * @author wangyu
 */
public class AuthenticationAuditorImpl extends ResultDataTransformer implements AuthenticationAuditor {

    @Resource
    private UserService userService;
    @Resource
    private AuthenticationLogger authenticationLogger;

    /**
     * 检查用户异常，无异常，清空错误次数，设置上次登录
     *
     * @param user 用户
     */
    @Override
    public Mono<Void> success(UserDetails user) {
        AdminUserDetails details = (AdminUserDetails) user;
        User updating = new User();
        updating.setId(details.getId());
        updating.setErrorCount(0);
        updating.setLastTime(LocalDateTime.now());
        return userService.updateSelectiveById(updating).then(authenticationLogger.success(user));
    }

    /**
     * 登录失败的回调
     * 当输入错误的账号密码，记录错误次数，否则只记录日志
     *
     * @param webFilterExchange 请求信息
     * @param exception         异常
     */
    @Override
    public Mono<Void> error(WebFilterExchange webFilterExchange, AuthenticationException exception) {
        return webFilterExchange.getExchange()
                .getFormData()
                .flatMap(data -> {
                    String username = data.getFirst("username");
                    Mono<User> userMono = userService.findByUsername(username).defaultIfEmpty(emptyUser(username));
                    // 当且仅当为凭据错误，尝试
                    if (exception instanceof BadCredentialsException) {
                        userMono = userMono.flatMap(user -> {
                            User updating = new User();
                            updating.setId(user.getId());
                            updating.setErrorCount(user.getErrorCount() + 1);
                            if (updating.getErrorCount() >= 5) {
                                updating.setStatus(UserStatus.LOCKED);
                            }
                            return userService.updateSelectiveById(updating);
                        });
                    }
                    return userMono.flatMap(user -> authenticationLogger.failure(user, exception));
                });
    }

    @Override
    public Mono<Void> logout(UserDetails userDetails) {
        return authenticationLogger.logout(userDetails);
    }

    /**
     * 构造空用户
     *
     * @param username 用户名
     * @return 结果
     */
    private User emptyUser(String username) {
        User user = new User();
        user.setUsername(username);
        user.setName("未知");
        return user;
    }

}
