package dev.flyfish.framework.beans.enums;

import dev.flyfish.framework.annotations.ConditionOn;
import dev.flyfish.framework.beans.meta.BeanPropertyType;
import dev.flyfish.framework.beans.meta.BeanValidation;
import dev.flyfish.framework.validation.annotations.IdCard;
import dev.flyfish.framework.validation.annotations.Money;
import dev.flyfish.framework.validation.annotations.Phone;
import dev.flyfish.framework.validation.annotations.UniqueField;
import dev.flyfish.framework.validation.enums.PhoneType;
import lombok.Getter;
import lombok.val;
import org.apache.commons.lang3.ArrayUtils;
import org.hibernate.validator.constraints.Length;
import org.hibernate.validator.constraints.Range;
import org.hibernate.validator.constraints.URL;
import org.springframework.core.annotation.MergedAnnotation;
import org.springframework.core.annotation.MergedAnnotations;

import jakarta.validation.constraints.*;
import java.lang.annotation.Annotation;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.function.BiConsumer;
import java.util.stream.Collectors;

/**
 * 校验候选元
 * 提供工厂方法进行构建
 *
 * @author wangyu
 */
@Getter
public enum ValidationCandidate {

    // 必填
    REQUIRED((annotation, validation) -> validation.setRequired(true), NotBlank.class, NotEmpty.class, NotNull.class),
    // 正则表达式
    PATTERN((annotation, validation) -> validation.setPattern(annotation.getString("regexp")), Pattern.class),
    // 必须为null
    NULL((annotation, validation) -> validation.setValidator("null"), Null.class),
    // 长度验证
    LEN((annotation, validation) -> {
        int min = annotation.getInt("min");
        int max = annotation.getInt("max");
        if (min == max) {
            validation.setLen(max);
        } else {
            validation.setMin(min).setMax(max);
        }
    }, Length.class, Size.class, Range.class),
    // 最小长度
    MIN((annotation, validation) -> validation.setMax(annotation.getInt("value")), Min.class),
    // 最大长度
    MAX((annotation, validation) -> validation.setMin(annotation.getInt("value")), Max.class),
    // 电子邮件
    EMAIL((annotation, validation) -> validation.setType("email"), Email.class),
    // URL地址
    URL_ADDRESS((annotation, validation) -> validation.setType("url"), URL.class),
    // 数字
    DIGITS((annotation, validation) -> validation.setType("number").setValidator("digits")
            .prop("integer", annotation.getInt("integer")).prop("fraction", annotation.getInt("fraction")),
            Digits.class),
    // 日期
    DATE((annotation, validation) -> validation.setValidator("date").prop("type", annotation.getType().getSimpleName()),
            Future.class, FutureOrPresent.class, Past.class, PastOrPresent.class),
    // 负数正数
    NATURE((annotation, validation) -> validation.setType("number").setValidator("nature")
            .prop("type", annotation.getType().getSimpleName()),
            Negative.class, NegativeOrZero.class, Positive.class, PositiveOrZero.class),
    // 货币
    MONEY((annotation, validation) -> validation.setValidator("currency"), Money.class),
    // 手机号
    PHONE((annotation, validation) -> validation.setValidator("phone").prop("type",
            annotation.getEnum("type", PhoneType.class).name().toLowerCase()), Phone.class),
    // 身份证号
    ID_CARD((annotation, validation) -> validation.setValidator("idCard"), IdCard.class),
    // 唯一字段
    UNIQUE_FIELD((annotation, validation) -> validation.setValidator("uniqueField")
            .prop("column", annotation.getString("value")), UniqueField.class);

    private final BiConsumer<MergedAnnotation<?>, BeanValidation> mapper;

    private final List<Class<? extends Annotation>> annotations;

    @SafeVarargs
    ValidationCandidate(BiConsumer<MergedAnnotation<?>, BeanValidation> mapper,
                        Class<? extends Annotation>... annotations) {
        this.mapper = mapper;
        this.annotations = Arrays.asList(annotations);
    }

    /**
     * 通过合并注解集合匹配
     *
     * @param annotations 注解集合
     * @return 结果
     */
    public static List<BeanValidation> produce(MergedAnnotations annotations, BeanPropertyType type) {
        return Arrays.stream(values())
                .flatMap(candidate -> candidate.annotations.stream()
                        .map(annotations::get)
                        .filter(MergedAnnotation::isPresent)
                        .map(annotation -> candidate.produce(annotation, type))
                ).collect(Collectors.toList());
    }

    /**
     * 生产校验规则
     *
     * @return 结果
     */
    public BeanValidation produce(MergedAnnotation<?> annotation, BeanPropertyType type) {
        BeanValidation validation = new BeanValidation()
                .setType(type.getValidType())
                .setMessage(annotation.getString("message"));
        Class<?>[] groups = annotation.getClassArray("groups");
        // 存在分组情况，为validation添加分组
        if (ArrayUtils.isNotEmpty(groups)) {
            val conditions = Arrays.stream(groups).map(clazz -> clazz.getAnnotation(ConditionOn.class))
                    .filter(Objects::nonNull)
                    .map(ConditionOn::expression)
                    .collect(Collectors.toList());
            validation.setConditions(conditions);
        }
        mapper.accept(annotation, validation);
        return validation;
    }
}
